#ifndef U2D_CORE_H
#define U2D_CORE_H

/* 

    umesh2dgui, a GUI for umesh2d grid generator 

    Copyright (C) 2003 Gabriele Dini Ciacci
    Copyright (C) 2004 Gabriele Dini Ciacci
    Copyright (C) 2005 Gabriele Dini Ciacci
   
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <math.h>


#define DEFAULT_TOLL 0.000001
#define PI 3.14159265358979
#define MAX_PRIM_STRING 1024
#define CIRCLE_APPROX 100000.0
#define ELLIPSE_APPROX 100000.0

#define PR_LINE 1
#define PR_CIRCLE 2
#define PR_ELLIPSE 3
#define PR_DATA 4

/* error codes */ 
#define SUCESS 0

/* file error codes */
#define NO_FILE -1
#define INVALID_FILE -2
#define NOT_A_FILE -3
#define NO_EXIST -4

/* other error codes */
#define NOMEM -10
#define UNKNOWN -256


/* Counterclockwise rotation matrix */
#define rot_matrixX(angle, X, Y)  ((X) * cos((angle) / 360.0 * 2.0 * PI) - (Y) * sin((angle) / 360.0 * 2.0 * PI))
#define rot_matrixY(angle, X, Y)  ((Y) * cos((angle) / 360.0 * 2.0 * PI) + (X) * sin((angle) / 360.0 * 2.0 * PI))

/* strctures for backgrids and steiners */
typedef struct {
  double x;
  double y;
} steiner;

typedef struct {
  double x;
  double y;
  double hdata;
} backgrid;

class Primitive {
 public:
  Primitive();
  virtual ~Primitive(){ };

  /* GEOMETRIC DATA */
  int spline;
  double x1;
  double y1;
  double x2;
  double y2;
  char primname[34];

  /* MANAGEMNET OF [i|s|h]DATA */
  double *data;
  /* NOTE: a data element is a triplet of [i|s|h]data */
  int maxdata; // maximum number of data elements for witch mem is allcoated
  int curdata; // current number of data elements memorized
  double getdata(int x, int y); // y from 0 to 2
  void setdata(int x, int y, double value); // y from 0 to 2
  int allocatedata(int ndata);

  /* INTERNAL MANAGEMNET */
  class Primitive *nextprimitive;
  class Primitive *prevprimitive;
  int type;
  int id;

  /* EXTERNAL MANAGEMENT */
  //  virtual int set_coord()=0;

  /* CLOSED OBJECT RELATED */
  /* Global progressive number of the primitive */
  unsigned int element;
  /* The object the primitive is part of.
   * An object is a group of one or more primitives.
   * 0 means unknown */
  unsigned int object;
  /* Progressive number of the primitive inside his object
   * 0 means it's unassigned */
  unsigned int object_part;
  /* Say if a primitive is part of the external boundary or not */
  bool isboundary;
  /* The number of the begin vertex and end vertex */
  int beginvertex;
  int endvertex;

  /* CLOCKWISENESS RELATED */
  virtual double calc_area(double ref_x, double ref_y)=0;
  virtual void reverse()=0;

 private:
  /* find_extremes() set the two extreme points where primitive start
   * and where it ends.
   * Does notting for PR_LINE naturally.
   * Each memeber define his funcion
   */
  virtual void find_extremes()=0;
};

class Line : public Primitive {
 public:
  Line() {
    type = PR_LINE;
  }
  ~Line(){  }
  int set_coord(int spli,\
		 double X1, double Y1,\
		 double X2, double Y2);
  double calc_area(double ref_x, double ref_y);
  void reverse();
  void find_extremes(){ };
};

class Circle : public Primitive {
 public:
  Circle() {
    type = PR_CIRCLE;
  }
  ~Circle() {  }
  int spline;
  double radius;
  double centerx;
  double centery;
  double startarc;
  double endarc;
  int set_coord(int spli, double rad,\
		 double centx, double centy,\
		 double sarc, double earc);
  double calc_area(double ref_x, double ref_y);
  void reverse();
 private:
  void find_extremes();
};

class Ellipse : public Primitive {
 public:
  Ellipse() {
    type = PR_ELLIPSE;
  }
  ~Ellipse() {  }
  int spline;
  double semix;
  double semiy;
  double centerx;
  double centery;
  double startarc;
  double endarc;
  int set_coord(int spli,\
		 double semx, double semy,\
		 double centx, double centy,\
		 double sarc, double earc);
  double calc_area(double ref_x, double ref_y);
  void reverse();
 private:
  void find_extremes();
};

class Data : public Primitive {
 public:
  Data() {
    type = PR_DATA;
    filename = NULL;
    points = NULL;
  }
  ~Data() {
    if(filename != NULL){
      free(this->filename);
    }
    if(points != NULL){
      free(this->points);
    }
  }
  char *filename;
  int npoints; /* number of points */
  double *points; /* points array */
  double trasx;
  double trasy;
  double rotation;
  int set_coord(const char* fname,\
		 double tx, double ty,\
		 double rot);
  double calc_area(double ref_x, double ref_y);
  void reverse();
 private:
  int allocatefilename(int length);
  void find_extremes();
};


/* Functions to find if a point is near another.
 * We use those to actualy consider really near points as the same point
 * The one with primitive arguments are not impelented yet
 */
bool isnear(double toll, double x1, double y1);
bool isnear(double toll, double x1, double y1, double x2, double y2);
bool isnear(double toll, Primitive *primX, Primitive *primY);

/* Function to calcualte the area of a triangle formed by a given line
   and a reference point */

double triangle_area(double ref_x, double ref_y,\
		     double x1, double y1,\
		     double x2, double y2);

class u2d_Core {
 public:
  u2d_Core();
  ~u2d_Core();

  /* PRIMITIVE MANGEMENT */
  Primitive *add_prim(int type);
  void del_prim(Primitive *oldprim);
  void move_insert_down_prim(Primitive *position, Primitive *insertion);
  int get_prim_number();
  Primitive *get_prim_list_start();
  Primitive *get_prim_list_end();
  char *get_prim_string(Primitive *curprim);
  Primitive *get_prim_by_label(const char *label);
  int get_vertex_n();

  /* Steiner and Backgrid points management TEMPORARILY PUBLIC WRITE GET METHODS*/
  int steiner_n;
  int backgrid_n;
  steiner *steiners;
  backgrid *backgrids;

  /* Assign objects numbers to primitives in the main list */
  void find_objects();

  int allocate_steiner(int number);
  int allocate_backgrid(int number);
  void set_steiner(int position, double x, double y);
  void set_backgrid(int position, double x, double y, double hdata);
  /* Debug */
  void debug_cb();

 private:
  /* PRIMITIVE MANGEMENT */
  Primitive *firstprimitive; 
  Primitive *lastprimitive;
  Primitive *new_prim(int type);
  void insert_down_prim(Primitive *position, Primitive *insertion);
  void remove_prim(Primitive *oldprim);
  char prim_string[1024];
  int elements;
  int vertex_n;
  int globalprimid; /* Consecutive number assigned to any
		     * generated primitive, can overflow */
  // in the future the id of the prim need to be managed much better.
  unsigned int last_object; /* This is the last closed object number */

  /* Steiner and Backgrid points management TEMPORARILY PUBLIC WRITE GET METHODS*/
  int steiner_max_n;
  int backgrid_max_n;
};

#endif
